#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2018 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Config Infrastructure Module (:mod:`tendril.utils.config`)
==========================================================

This module provides reusable infrastructure used by the tendril instance
configuration.

TODO Describe Architecture and Usage somewhere

"""

import os
import json
import importlib
import logging
from runpy import run_path
from distutils.util import strtobool
from tendril.utils.versions import get_namespace_package_names
from tendril.utils.files import yml

logger = logging.getLogger(__name__)


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except:
        return False


class ConfigElement(object):
    def __init__(self, name, default, doc, parser=None, masked=False):
        self.name = name
        self.default = default
        self.doc = doc
        self.parser = parser
        self.masked = masked
        self.source = None
        self.ctx = None

    def doc_render(self):
        return [self.name, self.doc, self.default,
                self.jsonable_value, self.source]

    @property
    def value(self):
        raise NotImplementedError

    @property
    def masked_value(self):
        """Return a masked version of the value of the option.

        If the option is not to be masked, the value is returned unchanged. 
        If the field is to be masked, the masked value is returned. The masking 
        algorithm is as follows:

        * If the value is not a string, it is returned unchanged.
        * If the value is a string, the first 8 characters (or 1/8 of the string, 
          whichever is shorter) are returned, followed by ellipses and the last 
          8 characters (or 1/8 of the string, whichever is shorter)

        (doc generated mostly by GitHub Copilot)
        Returns:
            str: The masked value of the field.
        """
        value = self.value
        if not self.masked:
            return value
        if not isinstance(value, str):
            return value

        v_len = len(value)
        m_len = int(min(v_len/8, 8))
        return f"{value[:m_len]}...{value[-m_len:]}"

    @property
    def jsonable_value(self):
        if is_jsonable(self.masked_value):
            return self.masked_value
        else:
            return str(self.masked_value)


def bool_parser(value):
    if not value:
        return False
    if isinstance(value, str):
        return strtobool(value)
    else:
        return bool(value)


class ConfigConstant(ConfigElement):
    """
    A configuration `constant`. This is fully specified in the core
    configuration module and cannot be changed by the user or the instance
    administrator without modifying the code.

    The value itself is constructed using ``eval()``.
    """
    @property
    def value(self):
        self.source = "hardcoded"
        return eval(self.default, self.ctx)


class ConfigOption(ConfigElement):
    """
    A configuration `option`. These options can be overridden
    by specifying them in the ``instance_config`` and
    ``local_config_overrides`` files.

    If specified in one of those files, the value should be
    the actual configuration value and not an expression. The
    default value specified here is used through ``eval()``.

    """
    @property
    def raw_value(self):
        try:
            rv = self.ctx['_environment_overrides'][self.name]
            self.source = 'environment_override'
            return rv
        except KeyError:
            pass

        try:
            rv = self.ctx['_local_config'][self.name]
            self.source = 'local_override'
            return rv
        except KeyError:
            pass

        try:
            rv = self.ctx['_instance_config'][self.name]
            self.source = 'instance_config'
            return rv
        except KeyError:
            pass

        try:
            if self.ctx['_external_configs']:
                rv = self.ctx['_external_configs'].get(self.name)
                self.source = 'external_config'
                return rv
        except ExternalConfigKeyError:
            pass

        try:
            rv = eval(self.default, self.ctx)
            self.source = 'default'
            return rv
        except SyntaxError:
            print("Required config option not set in "
                  "instance config : " + self.name)
            raise

    @property
    def value(self):
        """Get the value of the property.

        If a parser is defined, the value returned is the result of
        applying the parser to the raw value of the property. Otherwise,
        the raw value is returned.

        (doc generated mostly by GitHub Copilot)
        """
        if self.parser:
            if self.parser == bool:
                return bool_parser(self.raw_value)
            return self.parser(self.raw_value)
        else:
            return self.raw_value


class ConfigOptionConstruct(ConfigElement):
    def __init__(self, name, parameters, doc):
        self._parameters = parameters
        super(ConfigOptionConstruct, self).__init__(name, None, doc)

    @property
    def value(self):
        raise NotImplementedError


class ExternalConfigMissingError(Exception):
    def __init__(self, source, filetype):
        self.source = source
        self.filetype = filetype


class ExternalConfigFormatError(Exception):
    def __init__(self, source, filetype):
        self.source = source
        self.filetype = filetype


class ExternalConfigKeyError(Exception):
    def __init__(self, source, key):
        self.source = source
        self.key = key


class ConfigSourceDoesNotProvideKey(ExternalConfigKeyError):
    pass


class ConfigSourceDoesNotContainKey(ExternalConfigKeyError):
    def __init__(self, source, key, key_path):
        super(ConfigSourceDoesNotContainKey, self).__init__(source, key)
        self.key_path = key_path


class ConfigSourceBase(object):
    def get(self, key):
        raise NotImplementedError


class ConfigExternalSource(ConfigSourceBase):
    def __init__(self, path, keymap):
        self._path = path
        self._keymap: dict = keymap

    @property
    def path(self):
        return self._path

    def get(self, key):
        if key not in self._keymap.keys():
            raise ConfigSourceDoesNotProvideKey(self._path, key)
        return self._get(self._keymap[key])

    def _get(self, key_path):
        raise NotImplementedError


class ConfigExternalJSONSource(ConfigExternalSource):
    def __init__(self, path, keymap):
        super(ConfigExternalJSONSource, self).__init__(path, keymap)
        self._source = None
        self._load_external_config()

    def _load_external_config(self):
        """
        This function loads the external config file into a dictionary

        The external config file is a json file that contains some
        configuration parameters for the application.

        (doc generated mostly by GitHub Copilot)
        """
        if not os.path.exists(os.path.expandvars(self._path)):
            raise ExternalConfigMissingError(self._path, 'json')
        with open(os.path.expandvars(self._path), 'r') as f:
            self._source = json.load(f)

    def _get(self, key_path):
        """Return the value at the end of the key_path.

        key_path is a string of the form 'key1:key2:key3' where each key
        is a key in a dict. This function will return the value of key3
        in the dict d[key1][key2]. If key1, key2, or key3 do not exist,
        a ConfigSourceDoesNotContainKey error will be raised.

        (doc generated mostly by GitHub Copilot)
        """
        rval = self._source
        try:
            for crumb in key_path.split(':'):
                rval = rval.get(crumb)
        except KeyError:
            raise ConfigSourceDoesNotContainKey(self._source, None, key_path)
        return rval


class ConfigExternalSources(ConfigSourceBase):
    def __init__(self, path):
        super(ConfigExternalSources, self).__init__()
        self._path = path
        self._sources = []
        self._load_external_sources()

    def _load_external_sources(self):
        """
        This method loads the external config files from the given path.
        It iterates over each config and determines the format of the file.
        If the format is JSON, it will create a ConfigExternalJSONSource object
        and add it to the list of sources. If the file format is not supported,
        it will raise an ExternalConfigFormatError.

        (doc generated mostly by GitHub Copilot)
        """
        external_configs = yml.load(self._path)
        for config in external_configs:
            try:
                if config['format'] == 'json':
                    self._sources.append(
                        ConfigExternalJSONSource(config['path'], config['keymap'])
                    )
                else:
                    raise ExternalConfigFormatError(config['path'], config['filetype'])
            except ExternalConfigMissingError:
                pass

    def get(self, key):
        """Return the value associated with the key in the configured 
        external sources.

        Returns the value associated with the key in the first source
        that has the key. If the key is not found in any of the sources,
        raise ExternalConfigKeyError.

        (doc generated mostly by GitHub Copilot)

        Args:
            key (str): The key to search for.

        Returns:
            The value associated with the key.

        Raises:
            ExternalConfigKeyError: If the key is not found in any of
                the sources.
        """
        for source in self._sources:
            try:
                return source.get(key)
            except ExternalConfigKeyError:
                continue
        raise ExternalConfigKeyError(self._path, key)


class ConfigManager(object):
    """The ConfigManager class provides a consistent interface for accessing
    configuration information from a variety of sources. It is intended to be
    a singleton.

    (doc generated mostly by GitHub Copilot)
    """
    def __init__(self, prefix, legacy, excluded, appname=None):
        self._prefix = prefix
        self._excluded = excluded
        self.APPNAME = appname

        self._instance_config = None
        self._local_config = None
        self._external_configs: ConfigExternalSources = None
        self._environment_overrides = None
        self._modules_loaded = []
        self._legacy = None
        self._docs = []

        self._load_legacy(legacy)
        self._load_configs()

    def _check_depends(self, depends):
        for m in depends:
            if m not in self._modules_loaded:
                return False
        return True

    def _load_legacy(self, m_name):
        if not m_name:
            return
        logger.debug("Loading legacy configuration from {0}".format(m_name))
        self._legacy = importlib.import_module(m_name)

    @property
    def legacy(self):
        return self._legacy

    def _load_configs(self):
        """Load configuration elements from modules.

        This method will load configuration elements from modules in the 
        current namespace. It will respect the dependency order specified 
        in the depends attribute of each module while loading. If any 
        dependencies are not satisfied, the module will be skipped until 
        it is. If a module is skipped and there are no remaining modules
        to load, an error will be logged.

        (doc generated mostly by GitHub Copilot)
        """
        logger.debug("Loading configuration from {0}".format(self._prefix))
        modules = list(get_namespace_package_names(self._prefix))
        changed = True
        deadlocked = False
        while len(modules) and not deadlocked:
            if not changed:
                deadlocked = True
            changed = False
            remaining_modules = []
            for m_name in modules:
                if m_name in self._excluded:
                     continue
                m = importlib.import_module(m_name)
                if self._check_depends(m.depends):
                    logger.debug("Loading {0}".format(m_name))
                    m.load(self)
                    self._modules_loaded.append(m_name)
                    changed = True
                else:
                    if deadlocked:
                        logger.error("Failed loading {0}. Missing dependency."
                                     "".format(m_name))
                    remaining_modules.append(m_name)
            modules = remaining_modules

    def load_config_files(self):
        """ Loads the configuration from different sources to 
        populate the unified config object

        (doc generated mostly by GitHub Copilot)

        :return: None
        """
        if os.path.exists(self.INSTANCE_CONFIG_FILE):
            logger.info("Loading Instance Config from {0}"
                         "".format(self.INSTANCE_CONFIG_FILE))
            self._instance_config = run_path(self.INSTANCE_CONFIG_FILE)
        else:
            self._instance_config = {}

        if os.path.exists(self.LOCAL_CONFIG_FILE):
            logger.info("Loading Local Config from {0}"
                         "".format(self.LOCAL_CONFIG_FILE))
            self._local_config = run_path(self.LOCAL_CONFIG_FILE)
        else:
            self._local_config = {}

        try:
            self._environment_overrides = {}
            if self.ALLOW_ENVIRONMENT_OVERRIDES:
                if self.ENVIRONMENT_OVERRIDE_PREFIX:
                    for key, value in os.environ.items():
                        if key.startswith(self.ENVIRONMENT_OVERRIDE_PREFIX):
                            logger.info("Environment Config Override : {} : {}".format(key, value))
                            self._environment_overrides[key[len(self.ENVIRONMENT_OVERRIDE_PREFIX):]] = value
        except KeyError:
            pass

        if os.path.exists(self.EXTERNAL_CONFIG_SOURCES):
            logger.debug("Loading External Configuration Maps from {0}"
                         "".format(self.EXTERNAL_CONFIG_SOURCES))
            self._external_configs = ConfigExternalSources(self.EXTERNAL_CONFIG_SOURCES)

    @property
    def INSTANCE_CONFIG(self):
        return self._instance_config

    @property
    def LOCAL_CONFIG(self):
        return self._local_config

    @property
    def EXTERNAL_CONFIG(self):
        return self._external_configs

    @property
    def ENVIRONMENT_OVERRIDES(self):
        return self._environment_overrides

    def load_elements(self, elements, doc=''):
        """
        Loads the constants and/or options in the provided list into
        the config namespace.

        :param elements: `list` of :class:`ConfigConstant` or
                          :class:`ConfigOption` or
                          :class:`ConfigOptionConstruct`
        :return: None
        """
        _doc_part = []
        for element in elements:
            element.ctx = self.__dict__
            element.ctx['os'] = os
            setattr(self, element.name, element.value)
            _doc_part.append(element.doc_render())
        self._docs.append([_doc_part, doc])

    def instance_path(self, path):
        return os.path.join(self.INSTANCE_ROOT, path)

    @property
    def docs(self):
        return self._docs

    def doc_render(self):
        """Returns a dictionary of documentation for the options.

        The dictionary is keyed by the names of the option groups. 
        The value for each group is a dictionary keyed by the names 
        of the options in that group. 
        
        Each option value is a dictionary with the keys:
        
        doc
            The documentation string for the option.

        default
            The default value for the option.

        value
            The value of the option after all configuration files have been
            read.

        source
            The filename of the configuration file from which the option was
            read, or None if the option was not read from a configuration
            file.

        (doc generated mostly by GitHub Copilot)

        """
        rv = {}
        for section, name in self._docs:
            items = {}
            for oname, doc, default, masked_value, source in section:
                items[oname] = {'doc': doc, 'default': default,
                                'value': masked_value, 'source': source}
            rv[name] = items
        return rv

    def json_render(self):
        """Render the config as a dict suitable for JSON encoding.

        The structure is as follows:

            {
                'section1': {
                    'item1': {
                        'value': 'value1',
                        'source': 'source1',
                    },
                    'item2': {
                        'value': 'value2',
                        'source': 'source2',
                    },
                },
                'section2': {
                    'item1': {
                        'value': 'value1',
                        'source': 'source1',
                    },
                    'item2': {
                        'value': 'value2',
                        'source': 'source2',
                    },
                },
            }

        (doc generated mostly by GitHub Copilot)

        :returns: a dict of the config.
        """
        rv = {}
        for section, name in self._docs:
            items = {}
            for oname, doc, default, masked_value, source in section:
                items[oname] = {'value': masked_value, 'source': source}
            rv[name] = items
        return rv

    def log_render(self):
        """This method logs the rendered configuration. It loops through
        the sections and names of the configuration, and prints the
        value of each option, as well as its source (the file in which
        it was defined). The value is masked if the option is sensitive.

        (doc generated mostly by GitHub Copilot)
        """
        for section, name in self._docs:
            logger.info('--------------------------------')
            logger.info(f"{name.upper()} : ")
            for oname, _, _, masked_value, source in section:
                logger.info("    {0:30} :  {1}     ({2})".format(oname, masked_value, source))


def generate_constants(instance_name):
    config_constants_basic = [
        ConfigConstant(
            'INSTANCE_NAME',
            "'{}'".format(instance_name),
            'Name of the instance. Used to determine configuration and resource paths.'
        ),
        ConfigConstant(
            'INSTANCE_ROOT_CANDIDATES',
            """list([
                    os.path.join(os.path.expanduser('~'), '.{}'.format(INSTANCE_NAME), 'tendril'),
                    os.path.join('/etc', INSTANCE_NAME, 'tendril'),
                    os.path.join(os.path.expanduser('~'), '.tendril'),
            ])""",
            'Paths to search for the INSTANCE_ROOT. First available path will be used.'
        ),
        ConfigConstant(
            'INSTANCE_ROOT',
            "list(filter(os.path.exists, INSTANCE_ROOT_CANDIDATES))[0]",
            "Path to the instance root. Can be redirected if necessary"
            "with a file named ``redirect`` in this folder."
        ),
    ]

    config_constants_environment = [
        ConfigConstant(
            'ALLOW_ENVIRONMENT_OVERRIDES',
            "True",
            'Whether config options can be overridden from the environment.'
        ),
        ConfigConstant(
            'ENVIRONMENT_OVERRIDE_PREFIX',
            "'{}_'.format(INSTANCE_NAME.upper())",
            'Environment variable name prefix.'
        ),
    ]

    config_constants_redirected = [
        ConfigConstant(
            'INSTANCE_CONFIG_FILE',
            "os.path.join(INSTANCE_ROOT, 'instance_config.py')",
            'Path to the tendril instance configuration.'
        ),
        ConfigConstant(
            'LOCAL_CONFIG_FILE',
            "os.path.join(INSTANCE_ROOT, 'local_config_overrides.py')",
            'Path to local overrides to the instance configuration.'
        ),
    ]

    config_constants_external = [
        ConfigConstant(
            'EXTERNAL_CONFIG_SOURCES',
            "os.path.join(INSTANCE_ROOT, 'external_configs.yaml')",
            "Path to a yaml definition file mapping to external config sources."
        )
    ]
    return config_constants_environment, config_constants_basic, config_constants_redirected, config_constants_external


def install_config(manager, instance_name):
    config_constants_environment, \
    config_constants_basic, \
    config_constants_redirected, \
    config_constants_external = generate_constants(instance_name)

    manager.load_elements(config_constants_basic,
                          doc="Tendril Default Instance Root")

    manager.load_elements(config_constants_environment,
                          doc="Environment Variable Override Configuration")

    logger.info("Using Instance Root {}".format(manager.INSTANCE_ROOT))

    if os.path.exists(os.path.join(manager.INSTANCE_ROOT, 'redirect')):
        logger.info("Found instance redirect")
        with open(os.path.join(manager.INSTANCE_ROOT, 'redirect'), 'r') as f:
            manager.INSTANCE_ROOT = f.read().strip()
            logger.info("Using Redirected Instance Root {}".format(manager.INSTANCE_ROOT))

    manager.load_elements(config_constants_redirected,
                          doc="Tendril Configuration Paths")

    manager.load_elements(config_constants_external,
                          doc="External Configuration Sources")

    manager.load_config_files()
    



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
The Log Utils Module (:mod:`tendril.utils.log`)
===============================================

This module provides utilities to deal with logging systems. The intent of
having this module instead of using :mod:`logging` directly is to allow the
injection of various default parameters and options into all the loggers
used from a central place, instead of littering them throughout the
modules.

At present, this module does nothing that is overly useful, except for
being able to set the default log level for all modules simultaneously.

.. rubric:: Usage Example

>>> from tendril.utils import log
>>> logger = log.get_logger(__name__, log.DEFAULT)

"""

import os
import sys
import socket
import logging
from loguru import logger

#: Level for debug entries. High volume is ok
from logging import DEBUG   # noqa
#: Level for informational entires. Low volume
from logging import INFO  # noqa
#: Warnings only, which inform of possible failure
from logging import WARNING  # noqa
#: Errors only, which inform of high likelihood of failure
from logging import ERROR  # noqa
#: Critical Errors, things which should halt execution entirely
from logging import CRITICAL  # noqa


#: The default log level for all loggers created through this module,
#: unless otherwise specified at the time of instantiation.
DEFAULT = logging.INFO
_hostname = socket.gethostname()
_rename_modules = False
_names = {}
_source_maxlen = 15
identifier = None

logger_levels = {}
loggers = {}


def _time_fmt(config):
    """
    Return a time format string for use with the log formatter.

    The returned string is used by the log formatter to format the time
    portion of the log message.

    The time format string is determined by the configuration options
    LOG_COMPACT_TS and LOG_COMPACT_TS_READABLE.

    LOG_COMPACT_TS is a boolean option that determines whether the log
    formatter should use a compact format for the time portion of the log
    message.

    LOG_COMPACT_TS_READABLE is a boolean option that determines whether the
    log formatter should use a human-readable format for the time portion of
    the log message when the compact format is selected.

    (doc generated mostly by GitHub Copilot)
    
    Args:
        config: A Config object.

    Returns:
        A format string for use with the log formatter.
    """
    if config.LOG_COMPACT_TS:
        if config.LOG_COMPACT_TS_READABLE:
            return '{time:%m-%d %H%M.%S}'
        return '{time:%s}'
    return '{time:YYYY-MM-DD HH:mm:ss.SSS}'


def _hostname_fmt(config):
    """Add hostname to message if LOG_INCLUDE_HOSTNAME is set.

    If LOG_HOSTNAME_PREFIX is set, remove it from the hostname.

    (doc generated mostly by GitHub Copilot)
    """
    if config.LOG_INCLUDE_HOSTNAME:
        if config.LOG_HOSTNAME_PREFIX:
            return f' | {_hostname.removeprefix(config.LOG_HOSTNAME_PREFIX)}'
        else:
            return f' | {_hostname}'
    return ''


def _level_fmt(config):
    """Return a format string that will be used to format the log level.
    The log level can be displayed as a compact string, a compact icon, or a
    full name.
    
    (doc generated mostly by GitHub Copilot)
    """
    if config.LOG_COMPACT_LEVEL:
        return '{level.name:^1.1}'
    if config.LOG_COMPACT_LEVEL_ICON:
        return '{level.icon:^1}'
    return '{level: <8}'


def _source_fmt(config):
    """Return a format string that will format the message source.
    
    (doc generated mostly by GitHub Copilot)"""
    if config.LOG_COMPACT_SOURCE:
        return '{extra[name]}'
    return '{name}'


def _config(config):
    """ If the config LOG_COMPACT_SOURCE set to True, then we 
    patch each record to include a shorter version of the module name. 
    This compacted string is set in the 'extra' dictionary of the 
    record and the 'name' field is not modified. This is done to avoid 
    any potential side effects within the logging system.
    
    (doc generated mostly by GitHub Copilot)
    """
    if config.LOG_COMPACT_SOURCE:
        global _rename_modules
        global _source_maxlen
        _rename_modules = True
        _source_maxlen = config.LOG_COMPACT_SOURCE_MAXLEN
        patcher = lambda r: _shortname(r['name'], r['extra'])
        return patcher
    return


def apply_config(config=None):
    if not config:
        from tendril import config
    global DEFAULT
    global identifier
    global logger_levels
    logger_levels = {k: logging.getLevelName(v) for k, v in config.LOG_LOGGER_LEVELS.items()}
    DEFAULT = config.LOG_LEVEL
    logging.root.setLevel(config.LOG_LEVEL)

    patcher = _config(config)
    identifier = _hostname_fmt(config)

    fmt = "<green>" + _time_fmt(config) + identifier + "</green> | " \
          "<level>" + _level_fmt(config) + "</level> | " \
          "<cyan><i>" + _source_fmt(config) + "</i></cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> " \
          "- <level><n>{message}</n></level>"

    params = {
        'handlers': [{"sink": sys.stdout, "serialize": False, "format": fmt}],
    }
    if patcher:
        params['patcher'] = patcher
    logger.configure(**params)
    create_log_file(config)


def create_log_file(config):
    logdir = os.path.split(config.LOG_PATH)[0]
    if not os.path.exists(logdir):
        os.makedirs(logdir)

    fmt = "<green>" + _time_fmt(config) + _hostname_fmt(config) + "</green> | " \
          "<level>" + _level_fmt(config) + "</level> | " \
          "<cyan>" + _source_fmt(config) + "</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> " \
          "- <level><n>{message}</n></level>"

    logger.add(config.LOG_PATH, level="INFO", serialize=config.JSON_LOGS, enqueue=True,
               rotation="1 week", retention="14 days", format=fmt,
               catch=True, backtrace=True, diagnose=True)
    logging.info("Logging to: {}".format(config.LOG_PATH))


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def init():
    # intercept everything at the root logger
    logging.root.handlers = [InterceptHandler()]
    logging.root.setLevel(logging.INFO)

    # remove every other logger's handlers
    # and propagate to root logger
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = True

    # configure loguru
    logger.configure(handlers=[{"sink": sys.stdout, "serialize": False}])

    # logging.basicConfig(level=logging.DEBUG)
    silence = [
        logging.getLogger('watchdog.observers.inotify_buffer'),
        logging.getLogger('requests.packages.urllib3.connectionpool'),
        logging.getLogger('passlib.registry'),
        logging.getLogger('passlib.utils.compat'),
        logging.getLogger('appenlight_client.utils'),
        logging.getLogger('appenlight_client.timing'),
        logging.getLogger('appenlight_client.client'),
        logging.getLogger('appenlight_client.transports.requests'),
        logging.getLogger('pika.callback'),
        logging.getLogger('pika.channel'),
        logging.getLogger('pika.heartbeat'),
        logging.getLogger('pika.connection'),
        logging.getLogger('pika.adapters.base_connection'),
        logging.getLogger('pika.adapters.blocking_connection'),
        logging.getLogger('pika.adapters.select_connection'),
        logging.getLogger('urllib3.connectionpool'),
        logging.getLogger('matplotlib'),
        logging.getLogger('matplotlib.font_manager'),
        logging.getLogger('matplotlib.backends'),
        logging.getLogger('parso.python.diff'),
        logging.getLogger('parso.cache'),
    ]
    for external_logger in silence:
        external_logger.setLevel(logging.WARNING)
    logging.getLogger('pika.connection').setLevel(logging.ERROR)


def _shortname(name, extra):
    extra["name"] = _names.get(name, name)


_std_abbreviate = {'tendril': 't', 'libraries': 'lib'}
_never_abbreviate = ['db', 'config', 'mq']
_never_abbreviate_before = []
_never_abbreviate_after = []


def _tlen(parts):
    return sum([len(x) for x in parts]) + len(parts) - 1


def _recalculate_names():
    """This function is used to calculate the names of the
    modules that are used in the log messages. The names 
    thus calculated are stored in the _names dictionary and 
    used later via _shortname for each log message. This 
    dictionary contains the full name of the module as the 
    key, and the shortened name as the value.
    
    The shortened name is calculated by splitting the name 
    into a list of parts, and then abbreviating the parts 
    as needed using the tokens dictionary, which is 
    recalculated at every call to this function. 
    
    The parts are abbreviated if the length of the parts 
    list is greater than the maximum length, and if the 
    part is not in the list of parts that should never be 
    abbreviated.
    
    The parts of the name are abbreviated by abbreviating 
    the part to the shortest unique abbreviation.
    
    (doc generated mostly by GitHub Copilot)""" 
    global _names
    maxlen = _source_maxlen
    tokens = {}
    for name in _names.keys():
        parts = name.split('.')[:-1]
        for part in parts:
            current = tokens.get(part, {'abbrev': part, 'count': 0})
            current['count'] = current['count'] + 1
            tokens[part] = current
    for token in sorted(tokens.keys(), key=lambda x: tokens[x]['count'], reverse=True):
        if token in _std_abbreviate.keys():
            tokens[token]['abbrev'] = _std_abbreviate[token]
            done = True
        elif token in _never_abbreviate:
            tokens[token]['abbrev'] = token
            done = True
        else:
            done = False
        alen = 0
        while not done:
            alen = alen + 1
            abbrev = token[:alen]
            if abbrev not in [tokens[x]['abbrev'] for x in tokens.keys()]:
                tokens[token]['abbrev'] = abbrev
                done = True
    for name in _names.keys():
        parts = name.split('.')
        for idx, part in enumerate(parts[:-1]):
            if part in _never_abbreviate:
                continue
            try:
                if parts[idx + 1] in _never_abbreviate_before:
                    continue
            except IndexError:
                pass
            try:
                if parts[idx - 1] in _never_abbreviate_after:
                    continue
            except IndexError:
                pass
            if _tlen(parts) > maxlen:
                parts[idx] = tokens[part]['abbrev'] + '.'
            else:
                break
        _names[name] = '.'.join(parts)


def _register_name(name):
    global _names
    _names[name] = name
    if not _rename_modules:
        return
    _recalculate_names()


def get_logger(name, level=None):
    """
    Get a logger with the specified ``name`` and an optional ``level``.

    The levels from the python :mod:`logging` module can be used directly.
    For convenience, these levels are imported into this module's namespace
    as well, along with the :data:`DEFAULT` level this module provides.

    See python :mod:`logging` documentation for information about log levels.

    :param name: The name of the logger
    :type name: str
    :param level: Log level of the logger to be used.
                  Default : :data:`DEFAULT`.
    :type level: int
    :return: The logger instance
    """
    global loggers
    if name in loggers.keys():
        return loggers[name]
    built_logger = logging.getLogger(name)
    if level is not None:
        built_logger.setLevel(level)
    elif name in logger_levels.keys():
        built_logger.setLevel(logger_levels[name])
    else:
        built_logger.setLevel(DEFAULT)
    if name not in _names.keys():
        _register_name(name)
    loggers[name] = built_logger
    return built_logger


def set_logger_level(name, level):
    if name not in loggers.keys():
        return
    if isinstance(level, str):
        level = logging.getLevelName(level)
    loggers[name].setLevel(level)


def known_loggers():
    return {k: v.level for k, v in loggers.items()}


getLogger = get_logger

init()



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2016-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Collation Package Versions Module (:mod:`tendril.utils.versions`)
=================================================================

This module provides reusable infrastructure used for managing the landscape
and version information of the tendril namespace package collation.

TODO Describe Architecture and Usage somewhere

"""


from __future__ import print_function

import os
import rich
import pkg_resources
import pkgutil
import importlib


def get_namespace_package_names(namespace):
    ns_module = importlib.import_module(namespace)
    for _, name, _ in pkgutil.iter_modules(ns_module.__path__,
                                           ns_module.__name__ + '.'):
        yield name


def _namespace_primary_location(namespace, fpath):
    while os.path.split(fpath)[1] != namespace.split('.')[-1]:
        fpath = os.path.split(fpath)[0]
    return fpath


def get_namespace_package_locations(namespace):
    ns_package_names = get_namespace_package_names(namespace)
    ns_package_files = [pkgutil.get_loader(name).get_filename()
                        for name in ns_package_names]
    ns_package_locations = set([_namespace_primary_location(namespace, f)
                               for f in ns_package_files])
    return ns_package_locations


_prefixes = {'tendril'}
_packages = set()


def register_prefix(prefix):
    _prefixes.add(prefix)


def register_package(package):
    _packages.add(package)


def get_version(package):
    try:
        return pkg_resources.get_distribution(package).version
    except pkg_resources.DistributionNotFound:
        return 'Not Installed'


def get_versions(prefix):
    return sorted([(d.project_name, d.version)
                   for d in pkg_resources.working_set
                   if d.project_name.startswith(prefix)],
                  key=lambda x: x[0])


class FeatureUnavailable(Exception):
    def __init__(self, feature=None, provider=None):
        self._feature = feature
        self._provider = provider

    def __repr__(self):
        return "<FeatureUnavailable {0}>\nThis feature might be provided by " \
               "{1}.".format(self._feature, self._provider)


def versions():
    rv = {}
    for package in _packages:
        rv[package] = get_version(package)
    for prefix in _prefixes:
        val = {}
        _vers = get_versions(prefix)
        for name, version in _vers:
            val[name] = version
        rv[prefix] = val
    return rv


def main():
    try:
        from tendril import config
    except ImportError:
        pass
    rich.print_json(data=versions())


if __name__ == '__main__':
    main()






from pydantic import BaseModel


class TendrilTBaseModel(BaseModel):
    class Config:
        allow_population_by_field_name = True


class TendrilTORMModel(BaseModel):
    class Config:
        allow_population_by_field_name = True
        orm_mode = True





#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Docstring for test_utils_log
"""

from tendril.utils import log
import logging


def test_log_default():
    logger = log.get_logger("Default_Logger", log.DEFAULT)
    assert logger.name == "Default_Logger"
    assert logger.level == log.DEFAULT

    logger = log.get_logger("SP_Default_Logger", log.DEFAULT)
    assert logger.name == "SP_Default_Logger"
    assert logger.level == log.DEFAULT

    logger = log.get_logger("YA_Default_Logger")
    assert logger.name == "YA_Default_Logger"
    assert logger.level == log.DEFAULT



def test_log_other():
    logger = log.get_logger("Critical_Logger", log.CRITICAL)
    assert logger.name == "Critical_Logger"
    assert logger.level == logging.CRITICAL

    logger = log.get_logger("Error_Logger", log.ERROR)
    assert logger.name == "Error_Logger"
    assert logger.level == logging.ERROR

    logger = log.get_logger("Warning_Logger", log.WARNING)
    assert logger.name == "Warning_Logger"
    assert logger.level == logging.WARNING

    logger = log.get_logger("Info_Logger", log.INFO)
    assert logger.name == "Info_Logger"
    assert logger.level == logging.INFO

    logger = log.get_logger("Debug_Logger", log.DEBUG)
    assert logger.name == "Debug_Logger"
    assert logger.level == logging.DEBUG
